#ifndef _FWINONLINEARSOLVERS_H_
#define _FWINONLINEARSOLVERS_H_

#include <Newton.hpp>

#include "Helper/FWIDataContainerDomain.hpp"
#include "Helper/FWIDataContainerRange.hpp"
#include "Helper/NonlinearOptions.hpp"

class ForwardModel;

class DGSeismogram;

class FWIDataContainerRange;

class FWIDataContainerDomain;

typedef std::vector<FWIDataContainerRange> FWISubproblem;

typedef std::vector<std::vector<double>> NonlinearOutput;

class FWINonlinearSolver : public NonLinearSolver {
protected:
  int fwiverbose =0;
    Date Start;
    const std::string name;
    ForwardModel &FWD;
    double BacktrackingFactor{0.5};
    bool BacktrackingActivated{false};
public:
    FWINonlinearSolver(const char *_name, ForwardModel &FWD)
            : NonLinearSolver(nullptr, std::string(_name) + "_"), FWD(FWD),
               name(_name) {
        d = 0;
        Config::Get("FWI_BacktrackingFactor", BacktrackingFactor);
        Config::Get("FWI_BacktrackingActivated", BacktrackingActivated);
        Config::Get("fwiverbose",fwiverbose);
    }

  virtual NonlinearOutput operator()(const std::vector<FWIDataContainerRange> &s,
                                       FWIDataContainerDomain &, std::string namesuff) {
        Exit("Not implemented");
    }

    std::string Name() const override { return name; }

    void backtracking(const FWISubproblem &dataObs,
                      std::vector<double> &residualNorms,
                      const FWIDataContainerDomain &iterateMUpdate,
                      FWIDataContainerDomain &iterateNonlinear,
                      FWISubproblem &nonlinearEvaluation,
                      FWISubproblem &defect,
                      std::vector<int> subProb);

    void backtrack(const FWISubproblem &dataObs, std::vector<double> &residualNorms,
                   const FWIDataContainerDomain &iterateMUpdate,
                   FWIDataContainerDomain &iterateNonlinear,
                   FWISubproblem &nonlinearEvaluation,
                   FWISubproblem &defect, std::vector<int> subProb);
    virtual ~FWINonlinearSolver() {}
};

FWINonlinearSolver *GetFWINonlinearSolver(ForwardModel &FWD);

#endif
