//
// Created by cr on 26/03/23.
//

#ifndef FWI_FWIPROBLEMWRAPPER_HPP
#define FWI_FWIPROBLEMWRAPPER_HPP

#include <utility>

#include "FWIAbstractInterfaceDGTSVA.hpp"
#include "Problems/FWIAcousticInversionProb.hpp"
#include "VAcousticMatFromVec.hpp"

class FWIProblemWrapperVA2 : public ProblemVAcoustic {
    const FWIRhsVAcoustic &rhs;
    const FWIBndConditionsVAcoustic &bnd;
  std::unique_ptr<VAcousticMatFromVec> MatVec;
    const char *param;
    std::shared_ptr<FWIVAcousticInversionProb> fwd;
public:
    FWIProblemWrapperVA2(const FWIRhsVAcoustic &_rhs,
                        const FWIBndConditionsVAcoustic &_bnd,
                        std::unique_ptr<VAcousticMatFromVec> MatVec,
                        int _numL,
                        int _dim,
                        std::shared_ptr<FWIVAcousticInversionProb> fwd,
                        const char *_param = "native") : ProblemVAcoustic(_numL, _dim),
                                                         rhs(_rhs), bnd(_bnd),
                                                         MatVec(std::move(MatVec)),
                                                         param(_param), fwd(std::move(fwd)) {}

    virtual ~FWIProblemWrapperVA2() = default;

    bool RHS(double t) const override { return rhs.RHS(t); }

    bool HasExactSolution() const override { return bnd.exact(); }

    int getNrPointSrc() const override { return rhs.getNrPointSrc(); }

    Point getPointSrcLoc(int nr) const override { return rhs.getPointSrcLoc(nr); }

    double ForceP_i(double t, const Cell &C, const Point &z, int n) const override {
      return rhs.ForceP_i(t, C, z, n);
    }

    VectorField ForceV(double t, const Cell &C, const Point &z) const override {
      return rhs.ForceV(t, C, z);
    }

    void Force(VectorField &V_rhs, RVector &Force, double t, const Cell &C,
               const Point &z) const override {
      rhs.Force(V_rhs, Force, t, C, z);
    }

    double Dirichlet_i(double t, const Cell &, const Point &z, int i) const override {
        return bnd.Dirichlet_i(t, z, dim + i);
    }

    double Neumann(double t, const Cell &C, const Point z,
                   const VectorField &N) const override { return bnd.Neumann(t, C, z, N); }

    double ut(double t, const Cell &C, const Point &z, int k) const override {
        return bnd.ut(t, C, z, k);
    }

    virtual double u0(const Cell &C, const Point &x, int i) const override {
        return ut(0, C, x, i);
    }

    double Kappa(const Cell &C, const Point &x) const override;

    double Kappa_i(const Cell &C, const Point &x, int i) const override;

    double invKappa_i(const Cell &C, const Point &x, int i) const override {
        return 1.0 / Kappa_i(C, x, i);
    }

    double Tau_i(const Cell &c, const Point &x, int i) const override;

    virtual double invTau_i(const Cell &C, const Point &x, int i) const override {
        return 1.0 / Tau_i(C, x, i);
    }

    virtual double Rho(const Cell &C, const Point &x) const override;

    const char *Name() const override {
        mout << "yoyoy" << endl;
        return "Not implemented";
    }
   double alpha(const Cell &C,
                       const Point &x) const override {
    return fwd->alpha(C,x);
  }
   double alpha1() const override{
  return fwd->alpha1();
  }
   double alpha2() const override {
     return fwd->alpha2();
  }
};


class FWIProblemWrapper : public ProblemVAcoustic {
    const FWIRhsVAcoustic &rhs;
    const FWIBndConditionsVAcoustic &bnd;
    std::unique_ptr<Vector> MatVec;
    const char *param;
    std::shared_ptr<FWIVAcousticInversionProb> fwd;
public:
  FWIProblemWrapper(const FWIRhsVAcoustic &_rhs,
                        const FWIBndConditionsVAcoustic &_bnd,
                        std::unique_ptr<Vector> MatVec,
                        int _numL,
                        int _dim,
                        std::shared_ptr<FWIVAcousticInversionProb> fwd,
                        const char *_param = "native") : ProblemVAcoustic(_numL, _dim),
                                                         rhs(_rhs), bnd(_bnd),
                                                         MatVec(std::move(MatVec)),
                                                         param(_param), fwd(fwd) {}

    virtual ~FWIProblemWrapper() {}

    bool RHS(double t) const override { return rhs.RHS(t); }

    bool HasExactSolution() const override { return bnd.exact(); }

    int getNrPointSrc() const override { return rhs.getNrPointSrc(); }

    Point getPointSrcLoc(int nr) const override { return rhs.getPointSrcLoc(nr); }

    double ForceP_i(double t, const Cell &C, const Point &z, int n) const override {
      return rhs.ForceP_i(t, C, z, n);
    }

    VectorField ForceV(double t, const Cell &C, const Point &z) const override {
      return rhs.ForceV(t, C, z);
    }

    void Force(VectorField &V_rhs, RVector &Force, double t, const Cell &C,
               const Point &z) const override {
      rhs.Force(V_rhs, Force, t, C, z);
    }

    double Dirichlet_i(double t, const Cell &, const Point &z, int i) const override {
        return bnd.Dirichlet_i(t, z, dim + i);
    }

    double Neumann(double t, const Cell &C, const Point z,
                   const VectorField &N) const override { return bnd.Neumann(t, C, z, N); }

    double ut(double t, const Cell &C, const Point &z, int k) const override {
        return bnd.ut(t, C, z, k);
    }

    virtual double u0(const Cell &C, const Point &x, int i) const override {
        return ut(0, C, x, i);
    }

    double Kappa(const Cell &C, const Point &x) const override;

    double Kappa_i(const Cell &C, const Point &x, int i) const override;

    double invKappa_i(const Cell &C, const Point &x, int i) const override {
        return 1.0 / Kappa_i(C, x, i);
    }

    double Tau_i(const Cell &c, const Point &x, int i) const override;

    virtual double invTau_i(const Cell &C, const Point &x, int i) const override {
        return 1.0 / Tau_i(C, x, i);
    }

    virtual double Rho(const Cell &C, const Point &x) const override;

    const char *Name() const override {
        mout << "yoyoy" << endl;
        return "Not implemented";
    }
   double alpha(const Cell &C,
                       const Point &x) const override {
    return fwd->alpha(C,x);
  }
   double alpha1() const override{
  return fwd->alpha1();
  }
   double alpha2() const override {
     return fwd->alpha2();
  }
};


#endif  // FWI_FWIPROBLEMWRAPPER_HPP
