//
// Created by cr on 03.11.22.
//

#ifndef FWI_ABCHANDLER_HPP
#define FWI_ABCHANDLER_HPP

#include "DGAcousticElement.hpp"
#include "DGDiscretization.hpp"
#include "Helper/FWIHelpers.hpp"
#include "MeshesCreator.hpp"

class CosTaperConfig {
public:
  double deltad = 0.3;
  double vQ = 0.85;
  double vB = 0.2;
  double Qb = 1.1;
  double alpha1 = -1;
  double alpha2 = -2;
  CosTaperConfig(double alpha1_, double alpha2_, double deltad_ = 0.3,  double vQ_ = 0.85,
                 double vB_ = 0.2,  double Qb_ = 1.1) : alpha1(alpha1_),alpha2(alpha2_),
                                                        deltad(deltad_), vQ(vQ_),vB(vB_),Qb(Qb_){
    Qb = max(Qb_,alpha1/alpha2+100*Eps);
    //alpha1/alpha2 is the minimal Q that is possibile with this modelling
    mout <<  "Quality Factor set at BND: " << Qb << " Minimal in Modelling is: "
         <<alpha1/alpha2+100*Eps;
  }

};
vector<double> initializePMInftyVec(int size);
enum TAPERSTYLE {
  OLDTAPER = 0,
  COSTAPER = 1

};
class ABCConf {
  std::shared_ptr<CosTaperConfig> Cos = nullptr;
  vector<double> ABCPaddingLength;
  int taperstyle = OLDTAPER;
public:
  ABCConf()= default;
  vector<int> abcBndFlags;
  double distfactor = 1.0;
  void SetABCPaddingLength(const vector<double> &ABCPL){
    ABCPaddingLength = ABCPL;
  }
  [[nodiscard]] double GetABCPaddingLength(int i) const {
    return ABCPaddingLength[i];
  }
  [[nodiscard]] int getTaperStyle() const {
    return taperstyle;
  }
  [[nodiscard]] CosTaperConfig& GetCosConfig() const{
    return *Cos;
  }
  void SetTaperStyle(int taperstyle_,double alpha1 = -1, double alpha2 = -1) {
    taperstyle = taperstyle_;
    if (taperstyle_ == COSTAPER){
      if(alpha1 ==-1 && alpha2 == -1){
        MOUT("Cos Taper without damping");
      }
      double deltad = 0.3;
      double vQ = 0.85;
      double vB = 0.2;
      double Qb = 1.1;
      Config::Get("CosTaperEndCosPart",deltad);
      Config::Get("CosTaperRelativeValueAtEndOfCos",vQ);
      Config::Get("CosTaperRelativeVelocityAtBND",vB);
      Config::Get("CosTaperDampingAtBND",Qb);
      Cos = std::make_shared<CosTaperConfig>(alpha1,alpha2,deltad,vQ,vB,Qb);
    }
  }
};

using TransferABC = std::unordered_map<Point, std::pair<short, short>>;
using TransferToABCBoundary = std::unordered_map<Point, std::vector<std::pair<Point, short>>>;
using PointToPoint = std::unordered_map<Point,Point>;
class ABCHandler {
  void addCornerPoints(const vector<vector<double>> &outerBoundaryPoints, vector<Point> &Cmid, int dim, int nrbounds);
  TransferABC ABCToNonABC;
  TransferABC NonABCToABC;
  std::vector<double> dx;
  TransferToABCBoundary NonABCBoundaryToABC;
  int numL{};
  std::shared_ptr<EleCacheContainer> eleCacheContainerPML;
  ABCConf ConfigPML;
  FWIDiscretizationStack FWIDiscStackABC;
  vector<double> MeshBordersABC;
  vector<double> ProcBordersABC;
  vector<double> MeshBorders;
  vector<double> ProcBorders;
  bool initialized = false;

  struct CommDataRed{
    int sendProc;
    int ownerCellProc;
    Point CellPoint;
  };
  struct CommData{
    int sendProc;
    int borderProc;
    Point BorderPoint;
    Point ABCPoint;
  };
  void findBndFlags(const Mesh &, const vector<double> &, vector<int> &, vector<int> &);

  void findCellNrAndABCLength(const vector<double> &, vector<int> &, vector<int> &, vector<double> &, const vector<int> &,
                         int, std::vector<double> &);

  void initializeABCMesh(const Mesh &mesh, const vector<double> &MeshBordersI);


  void fillDiscStack(int deg, int numL, int paramComps);

  void initializeEleCache(const Vector &vec_, int numL);

public:
  PointToPoint ABCToBoundary;

  static void findMeshAndProcBorders(const Mesh &mesh, vector<double> &MeshBorders, vector<double> &ProcBorders);
  ABCHandler() = default;

  const ABCConf &GetConfig() const { return ConfigPML;}

  int GetNumL() const { return numL; }
  void CreatePointsOutsideOfInnerGrid(const Mesh &VecMesh, vector<std::pair<Point, vector<Point>>> &Cmids);

  int isOnProc(const Point &P, const vector<double> &VectorOfBoxes);

  void InitializeBoundaryCellToTransferNEW(
      TransferToABCBoundary &BondaryCellToTransfer,
                                           vector<std::pair<Point, vector<Point>>> &ExtrapolatedPoints,
                                           const vector<double> &Boundaries,
                                           const Vector &meshPML);

  void InitializeCellToTransfer(TransferABC &CellToTransfer, const Mesh &meshFrom, const Mesh& meshTo,
                                   const vector<double> &Boundaries);
  void Initialize(const Mesh &mesh, ABCConf Conf, int deg, int numL, int paramComps);

  FWIDiscretizationStack &GetDiscStack();

  std::shared_ptr<EleCacheType> GetEleCache() const;


  double GetDx() const {
    for (int i = 0; i < dx.size() - 1; ++i) {
      if (dx[i] != dx[i + 1]) {
        THROW("find usages of this and adapt it to differing dx");
      }
    }
    return dx[0];
  }

  bool isPML() const;

  const Vector &GetWaveVector() const;

  const Vector &GetMatVector() const;

  const vector<double> &GetProcBoundariesPML() const;

  const vector<double> &GetMeshBorders() const;

  const TransferABC &GetPMLToNonPML() const;

  const TransferABC &GetNonPMLToPML() const;

  const TransferToABCBoundary &GetNonPMLBoundaryToPML() const;

  double GetDistFactor() const {
    return ConfigPML.distfactor;
  }

  ~ABCHandler();
};
class ABCConfigAAO {
  std::shared_ptr<CosTaperConfig> Cos = nullptr;
  int taperstyle = OLDTAPER;
public:
  RectangularConfig RectConf;
  double distfactor = 1.0;
  ABCConfigAAO(){
    Config::Get("AAOPMLDistFactor", distfactor);
    Config::Get("InnerLeftBnd", RectConf.LeftBnd);
    Config::Get("InnerRightBnd", RectConf.RightBnd);
    Config::Get("InnerBottomBnd", RectConf.BottomBnd);
    Config::Get("InnerTopBnd", RectConf.TopBnd);
    Config::Get("InnerDimX", RectConf.dimX);
    Config::Get("InnerDimY", RectConf.dimY);
    int bnd = 0;
    RectConf.bndTop = bnd;
    RectConf.bndBottom = bnd;
    RectConf.bndLeft = bnd;
    RectConf.bndRight = bnd;
    RectConf.name = "Rect";

  }
  CosTaperConfig& GetCosConfig() const{
    return *Cos;
  }
  void SetTaperStyle(int taperstyle_,double alpha1 = -1, double alpha2 = -1) {
    taperstyle = taperstyle_;
    if (taperstyle_ == COSTAPER){
            if(alpha1 ==-1 && alpha2 == -1){
                  MOUT("Cos Taper without damping");
              }
      double deltad = 0.3;
      double vQ = 0.85;
      double vB = 0.2;
      double Qb = 1.1;
      Config::Get("CosTaperEndCosPart",deltad);
      Config::Get("CosTaperRelativeValueAtEndOfCos",vQ);
      Config::Get("CosTaperRelativeVelocityAtBND",vB);
      Config::Get("CosTaperDampingAtBND",Qb);
      Cos = std::make_shared<CosTaperConfig>(alpha1,alpha2,deltad,vQ,vB,Qb);
    }
  }
  int getTaperStyle() const{ return taperstyle;}
};
enum {
  LEFT =0,
  RIGHT = 1,
  DOWN= 2,
  UP =3,
};
class ABCHandlerAAO {
  double DistFactor = 1.0;
  std::vector<double> dx;
  vector<double> MeshBorders;
  vector<double> MeshBordersABC;
  vector<double> ProcBordersABC;
  LevelPair lev;
  TransferToABCBoundary NonABCBoundaryToABC;
  ABCConfigAAO Conf;
  int numL;
  vector<double> PMLPaddingLength;
public:
  explicit ABCHandlerAAO(const ABCConfigAAO &Conf, const Vector &vecfull,int numL);

  int isOnProc(const Point &P, const vector<double> &VectorOfBoxes) {
    int dim = (P.SpaceDim() + P.TimeDim());
    for (int i = 0; i < PPM->size(); ++i) {
      bool cont = true;
      for (int j = 0; j < dim; ++j) {
        bool tmp = VectorOfBoxes[i * 2 * dim + dim * j] <= P[j];
        bool tmp2 = VectorOfBoxes[i * 2 * dim + dim * j + 1] >= P[j];
        if (!tmp || !tmp2) cont = false;
      }
      if (cont) {
        return i;
      }
    }
    return -1;
  }
  struct CommData{
    int sendProc;
    int borderProc;
    Point BorderPoint;
    Point PMLPoint;
  };
  void InitializeBoundaryCellToTransferAAO(
      TransferToABCBoundary &BondaryCellToTransfer,
                                           vector<std::pair<Point, vector<Point>>> &ExtrapolatedPoints,
                                           const vector<double> &Boundaries,const Vector& meshfull);
  void SetABCZero(Vector& Vec){
    for (cell c = Vec.cells(); c!=Vec.cells_end(); c++){
      if( c()[0] > MeshBorders[0] && c()[0] < MeshBorders[1]
          && c()[1] > MeshBorders[2] && c()[1] < MeshBorders[3]){
        continue;
      }
      int d = Vec.Dof(Vec.Id(c()));
      for (int i =0; i < d; ++i) {
        Vec(c(),i) = 0.0;
      }
    }
  }
  void CommunicateBoundaryToPML(Vector &Vec) {
    ExchangeBuffer exBuffer;
    for (const auto &it: NonABCBoundaryToABC) {
      const auto &SendPoint = it.first;
      for (const auto &P: it.second) {
        short receiveProc = P.second;
        const auto &ReceivePoint = P.first;
        if (ReceivePoint == SendPoint) continue;
        exBuffer.Send(receiveProc) << ReceivePoint;
        bool found = Vec.find_cell(SendPoint) != Vec.cells_end();
        if (!found){
          THROW("Point" + to_string(SendPoint[0]) + " " + to_string(SendPoint[1])
                + " not found on proc " + to_string(PPM->Proc()));
        }
        int i = Vec.Id(SendPoint);
        for (int k = 0; k < Vec.Dof(i); ++k) {
          exBuffer.Send(receiveProc) << Vec(i, k);
        }
      }
    }
    Vec.CommunicateVector(exBuffer);
  }
  void increaseImpedance(Vector &Vec) {
    ExchangeBuffer exBuffer;
    for (const auto &it: NonABCBoundaryToABC) {
      const auto &SendPoint = it.first;
      for (const auto &P: it.second) {
        short receiveProc = P.second;
        const auto &ReceivePoint = P.first;
        if (ReceivePoint == SendPoint) continue;
        exBuffer.Send(receiveProc) << ReceivePoint;
        exBuffer.Send(receiveProc) << SendPoint;
      }
    }
    communicateImpedanceFactors(exBuffer,Vec);
  }

  void communicateImpedanceFactors(ExchangeBuffer &exBuffer, Vector& Vec) {
    exBuffer.Communicate();
    for (short q = 0; q < PPM->Size(Vec.CommSplit()); ++q) {
      while (exBuffer.Receive(q).size() < exBuffer.ReceiveSize(q)) {
        Point ReceivePoint;
        exBuffer.Receive(q) >> ReceivePoint;
        int i = Vec.Id(ReceivePoint);
        Point SendPoint;
        exBuffer.Receive(q) >> SendPoint;
        applydistanceFactors(Vec,SendPoint,ReceivePoint,i);
      }
    }
    exBuffer.ClearBuffers();
  }
  double OldTaper(double dist){ // this is basically an exponential taper in the velocity
    if (DistFactor * dist / dx[0] < 1) return 1;
    return DistFactor * dist / dx[0];
  }
  std::pair<double,double> calculateRelativeDistance(double distX,double distY,
                                                     vector<double> LayerSizes){
    double relDist = -infty;
    double Dist = -infty;
    if (abs(distX)>Eps) {
      int direc = distX < 0  ? LEFT : RIGHT;
      relDist = abs(distX) / LayerSizes[direc];
      Dist = LayerSizes[direc];
    }
    if (abs(distY) > Eps) {
      int direc = distY < 0  ? DOWN : UP;
      if (relDist <  abs(distY) / LayerSizes[direc]){
        relDist = abs(distY) / LayerSizes[direc];
        Dist = LayerSizes[direc];
      }
    }
    return {relDist,Dist};
  }
  void applydistanceFactors(Vector& Vec,
                            const Point &SendPoint, const Point &ReceivePoint, int i) {
    if (DistFactor < Eps) return;
    double distX = ReceivePoint[0] - SendPoint[0];
    double distY = ReceivePoint[1] - SendPoint[1];
    if (Conf.getTaperStyle() == OLDTAPER){
      double fac = OldTaper(max(abs(distY),abs(distX)));
      Vec(i,0) *= fac;
      for (int n =0; n <= numL; ++ n){
        Vec(i,1+n) /= fac;
      }
      return;
    }
    if(Conf.getTaperStyle() == COSTAPER){
      auto [relDist,Dist] = calculateRelativeDistance(distX,distY,PMLPaddingLength);
      double vQ = Conf.GetCosConfig().vQ;
      double vB = Conf.GetCosConfig().vB;
      double alpha1 = Conf.GetCosConfig().alpha1;
      double alpha2 = Conf.GetCosConfig().alpha2;
      double Qb = Conf.GetCosConfig().Qb;
      double Kappa0 = Vec(i, 1);
      double Kappa1 = Vec(i, 2);
      double taup = Kappa1 / Kappa0;
      double deltad = Conf.GetCosConfig().deltad;
      double C = relDist < deltad ? vQ + (1 - vQ) * cos(relDist / deltad * M_PI_2)
                                  : vQ + (vB - vQ) * (relDist - deltad) / (1 - deltad);
      if (numL == 0) {
        Vec(i, 0) /= C;
        Vec(i, 1) *= C;
      } else {
        double QM = (1.0 + taup * alpha1) / (taup * alpha2);
        double QT =
            pow(QM, 1.0 - relDist) * pow(Qb, relDist); // Exponential Taper from Q Value at
        double TauPML = 1.0 / (alpha2 * QT - alpha1);
        double quot1 = (1 + alpha1 * TauPML) / (1 + alpha1 * taup);
        double quot2 = (1.0 + numL * TauPML) / (1.0 + numL * taup);
        double RhoFac = sqrt(quot1 / quot2);
        double KappaFac = 1.0 / (sqrt(quot1 * quot2)); // these factors are supposed to ensure
        // the correct velocities as well as the correct impedance if the damping changes
        Vec(i, 0) *= RhoFac * 1.0 / C;
        Vec(i, 1) *= KappaFac * C;
        for (int n = 1; n <= numL; ++n) {
          Vec(i, 1 + n) = Vec(i, 1) * TauPML;
        }
      }

    }
    else Exit("Taper Not known");
  }
};
class ABCVectorFWI : public Vector {
protected:
  const ABCHandler &handler;

public:
  ABCVectorFWI(const ABCHandler &handler, const Vector &Vec) : Vector(0, Vec), handler(handler) {
    (*this).Clear();
  }

  static void transfer(const TransferABC &Transfer, const Vector &VecSend, Vector &VecRec);

  void PaddedToUnpadded(Vector &Vec);

  void UnpaddedToPadded(const Vector &Vec);

};



class ABCMatVectorFWI : public ABCVectorFWI {
  double OldTaper(double dist){ // this is basically an exponential taper in the velocity
    if (handler.GetDistFactor() * dist / handler.GetDx() < 1) return 1;
    return handler.GetDistFactor() * dist / handler.GetDx();
  }
  void applydistanceFactors(const Point &SendPoint,
                            const Point &ReceivePoint, int i,int j);

  double calculateRelativeDistance(double distX,double distY){
    double relDist = -infty;
    if (abs(distX)>Eps) {
      int direc = distX < 0  ? LEFT : RIGHT;
      relDist = abs(distX) / handler.GetConfig().GetABCPaddingLength(direc);
    }
    if (abs(distY) > Eps) {
      int direc = distY < 0  ? DOWN : UP;
      relDist = max(relDist,abs(distY) / handler.GetConfig().GetABCPaddingLength(direc));
    }
    return relDist;
  }
  void increaseImpedance();

  void communicateImpedanceFactors(ExchangeBuffer &exBuffer);

public:
  ABCMatVectorFWI(const ABCHandler &handler, const Vector &Unpadded, bool increaseImp);

  void CommunicateBoundaryToPML(const Vector &VecSend);
};

class ABCWaveVectorFWI : public ABCVectorFWI {
public:
  explicit ABCWaveVectorFWI(const ABCHandler &handler) : ABCVectorFWI(handler, handler.GetWaveVector()) {}
};

#endif  // FWI_ABCHANDLER_HPP
