//
// Created by cr on 14.11.22.
//

#ifndef FWI_FWIDATACONTAINERRANGE_HPP
#define FWI_FWIDATACONTAINERRANGE_HPP

#include "Observation/DGSeismogram.hpp"
#include "Vector.hpp"

class FWIDataContainerRange {
    std::unique_ptr<DGSeismogram> seismo = nullptr;
    std::unique_ptr<Vectors> Vecs = nullptr;
public:
    FWIDataContainerRange() = default;

    FWIDataContainerRange(const FWIDataContainerRange &cont);

    explicit FWIDataContainerRange(const DGSeismogram &s);

    explicit FWIDataContainerRange(const DGSeismogram &s, const Vectors &_vecs);

    explicit FWIDataContainerRange(const Vectors &_vecs);

    ~FWIDataContainerRange();

    bool isSeismo() const { return !(seismo == nullptr); }

    bool isWaveField() const { return !(Vecs == nullptr); }

    const DGSeismogram &getSeismogram() const { return *seismo; }

    const Vectors &getVectors() const { return *Vecs; }

    DGSeismogram &getSeismogramWrite() {
        if (!isSeismo()) {
            Exit("Seismogram not initialized!");
        } else {
            return *seismo;
        }
    }

    Vectors &getVectorsWrite() { return *Vecs; }

    std::unique_ptr<Vectors> &getVectorsPointer() { return Vecs; }

    FWIDataContainerRange &operator-=(const FWIDataContainerRange &s);

    FWIDataContainerRange &operator+=(const FWIDataContainerRange &s);

//    FWIDataContainerRange &operator=(const FWIDataContainerRange &s);

    FWIDataContainerRange &operator=(double val);

    FWIDataContainerRange &operator*=(double val);

    void resetWaveData();


};
typedef std::vector<FWIDataContainerRange> FWISubproblem;

FWISubproblem &operator*=(FWISubproblem &vec, double d);

FWISubproblem &operator-=(FWISubproblem &vec1, FWISubproblem &vec2);

FWISubproblem &operator+=(FWISubproblem &vec1, FWISubproblem &vec2);

#endif //FWI_FWIDATACONTAINERRANGE_HPP
