#include "FWIMppHelpers.hpp"

double zKPrimeTauPositive(int numL, double alpha, double rhoK, double vK, double taupK, double rhoDK, double vDK,
                          double taupDK) {
    double zK = zKTauPositive(numL, alpha, rhoK, vK, taupK);
    if (numL == 0) {
        return (rhoDK + vDK) * zK;
    } else {
        double expTaupK = exp(taupK);
        double taupKFac = expTaupK * (numL - alpha)
                          / (2.0 * (1 + numL * expTaupK) * (1.0 + alpha * expTaupK));
        return (rhoDK + vDK + taupKFac * taupDK) * zK;
    }
}

double zKTau(int numL, double alpha, double rhoK, double vK, double taupK) {
    if (numL == 0) {
        return rhoK * vK;
    } else {
        return rhoK * vK * sqrt((1.0 + numL * taupK) / (1 + alpha * taupK));
    }
}

double
zKPrimeTau(int numL, double alpha, double rhoK, double vK, double taupK, double rhoDK, double vDK, double taupDK) {
    double zK = zKTau(numL, alpha, rhoK, vK, taupK);
    if (numL == 0) {
        return ((rhoDK / rhoK) + (vDK / vK)) * zK;
    } else {
        double taupKFac =
                (numL - alpha) / (2.0 * (1 + numL * taupK) * (1.0 + alpha * taupK));
        return ((rhoDK / rhoK) + (vDK / vK) + taupKFac * taupDK) * zK;
    }
}

double zKTauPositive(int numL, double alpha, double rhoK, double vK, double taupK) {
    if (numL == 0) {
        return exp(rhoK) * exp(vK);
    } else {
        double expTaupK = exp(taupK);
        return exp(rhoK) * exp(vK)
               * sqrt((1.0 + numL * expTaupK) / (1 + alpha * expTaupK));
    }
}

std::vector<Point> PointsOnLine(vector<Point> &waypoints,
                                vector<int> &nrOfPointsOnLine) {
    unsigned int waypoint_count = waypoints.size();


    std::vector<Point> PointsOnLine;
    Point start = waypoints[0];
    for (int i = 1; i < waypoint_count; ++i) {
        Point end = waypoints[i];
        int count = nrOfPointsOnLine[i - 1];
        if (count == 1 && norm(start-end) < Eps){
          PointsOnLine.push_back(end);
          continue;
        }else if(count == 1 && norm(start-end) > Eps){
          THROW("Only one point on line but start and end are not the same!")
        }
        for (int j = 0; j < count; ++j) {
            double lambda = double(j) / (count-1);
            PointsOnLine.push_back((1.0 - lambda) * start + lambda * end);
        }
        start = end;
    }
    return PointsOnLine;
}

void ReadWaypoints(const string& Wpname, int& WpCount, vector<Point>& WpPoints,
                   std::vector<int>& nrPointsBetween) {
    int tmp = WpCount;
    Config::Get("FWI" + Wpname + "Waypoints", WpCount);
    if (tmp == WpCount) {
        if (nrPointsBetween.empty()) nrPointsBetween = std::vector<int>(WpCount - 1, 0);
        if (WpPoints.empty()) WpPoints = std::vector<Point>(WpCount, Infty);
    } else {
        int sizeold = WpPoints.size();
        nrPointsBetween.resize(WpCount - 1);
        WpPoints.resize(WpCount);
        if (sizeold < WpCount) {
            for (int i = sizeold; i < WpCount; ++i) {
                WpPoints.at(i) = Infty;
            }
            for (int i = sizeold; i < WpCount - 1; ++i) {
                nrPointsBetween.at(i) = 0;
            }
        }
    }
    for (int i = 0; i < WpCount; ++i) {
        mout << ("FWI" + Wpname + "Count" + std::to_string(i)).c_str() << endl;
        if (i != WpCount - 1) Config::Get("FWI" + Wpname + "Count" + std::to_string(i), nrPointsBetween[i]);
        if (i < WpCount) {
            Config::Get("FWI" + Wpname + "Waypoint" + std::to_string(i), WpPoints[i]);
            if (WpPoints[i] == Infty) THROW("Failed to specify ReceiverWaypoint " + std::to_string(i));
        }
    }
    if (WpCount < 2) Exit("At least 2 waypoints needed!" + Wpname);
}
