//
// Created by cr on 25/01/24.
//
#include "FWIObservationOperatorSpaceDGTS.hpp"

void FWIObservationOperatorSpacePoint::initalizeShapeValues(const ObservationSpecification &spec,
                                                            const Mesh &b) {
  for (cell c = b.cells(); c != b.cells_end(); ++c) {
    for (const auto &receiver: spec) {
      if (c.PointInCell(receiver)) {
        SpaceReceiver2SpaceSupport[receiver].push_back(c());
        auto elem = *EleCache->find((*c)())->second;
        int nV = elem.V_dimension();
        int nP = elem.shape_size();
        RVector PV(0.0, nP);
        Point lz = (*c).GlobalToLocal(receiver);
        for (int i = 0; i < nP; ++i) {
          PV[i] = elem.evaluateShape(lz, i);
        }
        PVals[receiver].emplace(c(), PV);
        if (measureComp == 1) {
          std::vector<VectorField> VFV(nV, zero);
          for (int i = 0; i < nV; ++i) {
            VFV[i] = elem.evaluateVelocityShape(lz, i);
          }
          VFVals[receiver].emplace(c(), VFV);
        }
      }
    }
  }
}

void FWIObservationOperatorSpacePoint::SetSpaceValue(const FWIState &Measure, const Point &Receiver,
                                                     const std::__cxx11::list<Point> &MeasureCells,
                                                     Vector &b) const {
  auto PValsAtRec = PVals.at(Receiver);
  auto cell_count = ReceiverToCellCount.at(Receiver);
  for (const auto &MeasureCell: MeasureCells) {
    auto elem = *EleCache->find(MeasureCell)->second;
    int p_dim = elem.p_dimension();
    int nP = elem.shape_size();
    const RVector &PV = PValsAtRec.at(MeasureCell);
    for (int i = 0; i < nP; ++i) {
      for (int n = 0; n < p_dim; ++n) {
        b(elem[0], (b.dim() + n) * nP + i) += Measure[b.dim() + n] * PV[i] * (1.0 / cell_count);
      }
    }
    if (measureComp == 1) {
      const auto &VF = VFVals.at(Receiver).at(MeasureCell);
      for (int i = 0; i < nP; ++i) {
        for (int j = 0; j < b.dim(); ++j) {
          b(elem[0], j * nP + i) += Measure[j] * VF[i][j] * (1.0 / cell_count);
        }
      }
    }
  }
}



FWIState FWIObservationOperatorSpacePoint::GetSpaceValue(const Vector &U, const Point &Receiver,
                                                         const std::list<Point> &MeasureCells) const {
  FWIState Measure(stateComps);
  auto PValsAtRec = PVals.at(Receiver);
  auto cell_count = ReceiverToCellCount.at(Receiver);
  for (const auto &MeasureCell: MeasureCells) {
    auto elem = *EleCache->find(MeasureCell)->second;
    int p_dim = elem.p_dimension();
    int nP = elem.shape_size();
    const RVector &PV = PValsAtRec.at(MeasureCell);
    for (int i = 0; i < nP; ++i) {
      for (int n = 0; n < p_dim; ++n) {
        Measure[U.dim() + n] += U(elem[0], (U.dim() + n) * nP + i) * PV[i];
      }
    }
    if (measureComp == 1) {
      const auto &VF = VFVals.at(Receiver).at(MeasureCell);
      for (int i = 0; i < nP; ++i) {
        for (int j = 0; j < U.dim(); ++j) {
          Measure[j] += U(elem[0], j * nP + i) * VF[i][j];
        }
      }
    }
  }
  return Measure * (1.0 / cell_count);
}