//
// Created by cr on 24.11.22.
//

#ifndef FWI_WAYPOINTSOURCES_HPP
#define FWI_WAYPOINTSOURCES_HPP

#include "Helper/FWIMppHelpers.hpp"
#include "Point.hpp"
#include "lib6_app/seismogram/ObservationSpecification.hpp"
class SourcesPositions {
protected:
  int shot = -1;
  int numberOfSourcesPerShot = 1;
  std::vector<std::vector<Point>> shotLocations;
public:
  SourcesPositions(const std::vector<std::vector<Point>> &shotLocations,
                   int numberOfSourcesPerShot = 1)
      : shotLocations(shotLocations), numberOfSourcesPerShot(numberOfSourcesPerShot) {
    mout.PrintInfo("Sources", 1, PrintInfoEntry("Source Locations:", shotLocations));
  }
  SourcesPositions() = default;
  Point GetShotLocation(int nr) const {
    return shotLocations[shot][nr];
  }
  int GetNrSrces() const {
    return shotLocations[shot].size();
  }

  int GetShotCount() const {
    return shotLocations.size();
  }
  void SetShot(int shot_) {
    shot = shot_;
  }

  int GetShot() const {
    return shot;
  }
};
class ReadSources : public SourcesPositions {
public:
  ReadSources() = default;
  explicit ReadSources(const std::string &fn, int shotStart, int shotEnd) : SourcesPositions(){
    std::ifstream file(fn);
    if (!file.good()) {
      Exit("No text-file found at: " + fn)
    }
    std::string line;
    std::vector<std::vector<Point>> shotLocationsTmp;

    while (getline(file, line)) {
      trim(line);
      if (line.length() == 0) {
        continue;
      }
      size_t equalPos = line.find(',');
      if (equalPos == string::npos) {
        continue;
      }
      string x = line.substr(0, equalPos);
      trim(x);
      string y = line.substr(equalPos + 1);
      trim(y);
      shotLocationsTmp.emplace_back(1,Point(stod(x),stod(y)));
    }
    if (shotStart > shotEnd) Exit("error");
    if (shotLocationsTmp.size() < shotEnd) Exit("error");
    mout << "Sources that were read: " << endl;
    for(int i = shotStart ; i < shotEnd;++i){
      mout << shotLocationsTmp[i] << endl;
      shotLocations.push_back(shotLocationsTmp[i]);
    }
  }

};


class WaypointSources : public SourcesPositions {
    void CheckOldOptions();
    void initialize(vector<Point>& waypoints, vector<int>& SourceCounts);
public:
    WaypointSources();

    WaypointSources(std::vector<Point> waypoints,
                    std::vector<int> SourceCounts);
};

#endif //FWI_WAYPOINTSOURCES_HPP
