//
// Created by cr on 21/12/22.
//

#include "FWITestLinearizedMatrix.h"

#include <experimental/random>

#include "Helper/NormCalculation.hpp"

std::map<Point, const Cell&> GetOrderedCells(cell start, cell end){
  std::map<Point, const Cell&> result;
  for(cell c=start; c!= end; c++){
    result.emplace(c(), *c);
  }
  return result;
}

TEST_P(FWITestLinearizedMatrix, AdditionTest){
  Vector Rand(*Mat);
  Vector U(*discStack->wV);
  Rand.Clear();
  std::experimental::reseed(42);
  for (auto& [point, c] : GetOrderedCells(Rand.cells(), Rand.cells_end())) {
    auto r = 2 * double(std::experimental::randint(0, 10000) / 10000.0);
    auto r2 = 2 * double(std::experimental::randint(0, 10000) / 10000.0);
    auto r3 = 2 * double(std::experimental::randint(0, 10000) / 10000.0);
    (*Mat)(c(), 0) = r;
    (*Mat)(c(), 1) = r2;
    (*Mat)(c(), 2) = r3;
  }
  auto assnorm = GetAssemble(*Mat);
  Matrix DampingPart(U);
  Matrix FluxPart(U);
  Matrix FullMatrix(U);
  assnorm->DampingMatrix(DampingPart);
  assnorm->FluxMatrixOnly(FluxPart);
  assnorm->SystemMatrix(FullMatrix);
  DampingPart += FluxPart;
  FullMatrix += (-1.0) * DampingPart;
  EXPECT_NEAR(FullMatrix.norm(),0.0,1e-16);
}
INSTANTIATE_TEST_SUITE_P(AdditionTest, FWITestLinearizedMatrix, Values(
    TestParams{"native",0,MASS_MATRIX, {0, 0, 0, 0, 0, 0}})
    );
TEST_P(FWITestLinearizedMatrix, FWITestLinearizedMatrix) {
  Vector Rand(*Mat);
  Vector U(*discStack->wV);
  Rand.Clear();
  std::experimental::reseed(42);
  for (auto& [point, c] : GetOrderedCells(Rand.cells(), Rand.cells_end())) {
    auto r = 2 * double(std::experimental::randint(0, 10000) / 10000.0);
    auto r2 = 2 * double(std::experimental::randint(0, 10000) / 10000.0);
    auto r3 = 2 * double(std::experimental::randint(0, 10000) / 10000.0);
    (*Mat)(c(), 0) = r;
    (*Mat)(c(), 1) = r2;
    (*Mat)(c(), 2) = r3;
  }
  FWILinearizedAssembler ass(prob, *eleCache->Get(), *Mat);
  auto assnorm = GetAssemble(*Mat);
  for (auto& [point, c] : GetOrderedCells(Rand.cells(), Rand.cells_end())) {
    auto r = 2 * double(std::experimental::randint(0, 10000) / 10000.0);
    Rand(c(), GetParam().param) = r;
  }
  Rand *= 1.0 / nc::L2MaterialNorm(Rand, eleCache->Get(), reconlist);
  Matrix Prime(U);
  Matrix M(U);
  Vector MatPlusRand(*Mat + Rand);
  GetMatrix(assnorm,M,GetParam().matrix);
  GetLinearizedMatrix(ass,Rand,Prime,GetParam().matrix);
  Matrix MPlusRand(M);
  MPlusRand = 0.0;
//  MOUT(Prime);
//  MOUT(M);
  auto assnormPlusRand = GetAssemble(MatPlusRand);
  GetMatrix(assnormPlusRand,MPlusRand,GetParam().matrix);
//  MOUT(MPlusRand);
  std::vector<double> quotients(6, 0.0);
  for (int i = 0; i < 6; ++i) {
    Matrix MPlusRand(M);
    double fac = pow(10, -i);
    MatPlusRand = *Mat;
    MatPlusRand += fac * Rand;
    auto assnormPlusRand = GetAssemble(MatPlusRand);
    GetMatrix(assnormPlusRand,MPlusRand,GetParam().matrix);
    MPlusRand += (-1.0) * M;
//    MOUT(MPlusRand);
    MPlusRand += (-fac) *Prime;
//    MOUT(MPlusRand);
    quotients[i] = MPlusRand.norm()/fac;
  }
  MOUT(quotients);
  EXPECT_VECTOR_NEAR(quotients, GetParam().values, max(quotients[0] * 1e-7,1e-10));
}
INSTANTIATE_TEST_SUITE_P(FWITestLinearizedMatrix, FWITestLinearizedMatrix, Values(
    TestParams{"native",0,MASS_MATRIX, {0, 0, 0, 0, 0, 0}},
    TestParams{"native",1,MASS_MATRIX, {22.109882, 14.237205, 3.1700744, 0.36189663, 0.036710521,
                                        0.0036763444}},
    TestParams{"native",2,MASS_MATRIX, {61.136744, 49.591937, 17.289952, 2.3023656, 0.23813514,
                                        0.023895493}},
    TestParams{"native",2,DAMPING_MATRIX, {3841.3349, 3115.9533, 1086.3597, 144.66189, 14.962472, 1.5013981}},
    TestParams{"taup",0,MASS_MATRIX, {32954.718284568997, 32028.742615239782, 25094.295839286307,
                                      7928.8218185107989, 1011.2793, 103.99237}},
    TestParams{"taup",1,MASS_MATRIX, {82379.560659440976, 62740.695595155397, 17083.167078519928,
                                      2041.0486291673767, 208.11016, 20.851881}},
    TestParams{"taup",2,MASS_MATRIX, {19909.763289753446, 10353.917154006049, 1785.3430469132998,
                                      192.47927, 19.399476,1.9414763}},
     TestParams{"taup",0,DAMPING_MATRIX, {1165343.4, 1131428.9, 886437.03, 280079.45,  35722.651, 3673.4492}},
    TestParams{"taup",1,DAMPING_MATRIX, {5170541.1, 3937922.6, 1072229, 128106.97, 13062.091, 1308.774}},
    TestParams{"taup",2,DAMPING_MATRIX, {1250967.3, 650555.8, 112176.41, 12093.83, 1218.905, 121.98655}},
    TestParams{"taup_positive",0,MASS_MATRIX, {0.18871896, 0.015032206, 0.0014720252, 0.00014689676,
                                               1.4686625e-05,1.4686341e-06}},
    TestParams{"taup_positive",1,MASS_MATRIX, {0.046756905, 0.006844487, 0.00071500229,
                                               7.1817585e-05, 7.1849442e-06,7.1852584e-07}},
    TestParams{"taup_positive",2,MASS_MATRIX, {0.015060111, 0.0011947275, 0.00011700487,
                                               1.1676372e-05, 1.1673966e-06, 1.1673685e-07}},
    TestParams{"taup_positive",0,DAMPING_MATRIX, {0.27987723, 0.033652253, 0.0034325371, 0.00034393923, 3.4400788e-05, 3.4401457e-06}},
    TestParams{"taup_positive",1,DAMPING_MATRIX, {0.93587548, 0.13171988, 0.013699788, 0.0013754518, 0.00013760011, 1.3760545e-05}},
    TestParams{"taup_positive",2,DAMPING_MATRIX, {0.13805148, 0.016295158, 0.0016587577,
                                                  0.00016617322, 1.6620304e-05, 1.6620585e-06}}
                                                ));
int main(int argc, char **argv) {
  return MppTest(
      MppTestBuilder(argc, argv).
          WithoutDefaultConfig().
          WithScreenLogging().
          WithConfPath("../").
          WithGeoPath("../conf/geo/").
          WithPlotPath("data/vtu/").
          WithPPM()
  ).RUN_ALL_MPP_TESTS();
}