//
// Created by cr on 23.08.22.
//

#ifndef FWI_DGVACOUSTICASSEMBLECOMBINED_HPP
#define FWI_DGVACOUSTICASSEMBLECOMBINED_HPP

#include <utility>

#include "Assemble.hpp"
#include "DGVAcousticMatrixAssembler.hpp"
#include "DGVAcousticPlotter.hpp"
#include "DGVAcousticRHSAssembler.hpp"
// #include "Assembler/FWIRhsAssemblerLinearized.hpp"

struct AcousticErrorsDG {
  double l1 = 0.0;

  double l2 = 0.0;

  double inf = 0.0;
};

struct AcousticNormsDG {
  double l1 = 0.0;

  double l2 = 0.0;

  double energy = 0.0;
};
typedef std::pair<double, double> MinMaxPDG;

class DGVAcousticAssembleCombined : public ILinearTimeAssemble {
  const std::unordered_map<Point, DGAcousticElement *> *EleCache;
  bool useCollectTweak = false;
  Vector *u0 = nullptr;

private:
  std::shared_ptr<DGVAcousticPlotter> plotter;
  std::shared_ptr<DGVAcousticMatrixAssembler> assembler;
  std::shared_ptr<DGVAcousticRHSAssembler> rhsAssembler;

public:
  DGVAcousticAssembleCombined(std::shared_ptr<DGVAcousticPlotter> plotter,
                              std::shared_ptr<DGVAcousticMatrixAssembler> assembler,
                              std::shared_ptr<DGVAcousticRHSAssembler> rhsAssembler, double t0,
                              double T, double dt)
      : ILinearTimeAssemble(t0, T, dt), plotter(std::move(plotter)),
        assembler(std::move(assembler)), rhsAssembler(std::move(rhsAssembler)) {
    Config::Get("UseCollectTweak", useCollectTweak);
  }

  void SetRhsAssembler(std::shared_ptr<DGVAcousticRHSAssembler> rhsAssembler_) {
    rhsAssembler = std::move(rhsAssembler_);
  }

  void SetEleCache(const std::unordered_map<Point, DGAcousticElement *> *eleCache);

  ~DGVAcousticAssembleCombined() override;

  const char *Name() const override;

  void Initialize(Vector &u) const override;

  void RHS(double t, Vector &rhs) const override;

  void MassMatrix(Matrix &massMatrix) const override;

  void SystemMatrix(Matrix &systemMatrix) const override;

  void FinishTimeStep(const Vector &u) override;

  void DefineInitialValue(Vector *u) { u0 = u; }

  void SetInitialValue(Vector &u) override;

  void PrintInfo() const override;

  void PrintIteration(const Vector &u) const override;

  void PlotIteration(const Vector &u) const override;

  void InterpolatedExactSolution(Vector &u) const;

  AcousticErrorsDG ComputeErrors(const Vector &u) const;

  AcousticNormsDG ComputeNorms(const Vector &u) const;

  double Energy(const Vector &u) const;

  double Energy_local(const Vector &u, const DGAcousticElement &elem, const cell &c) const;

  std::pair<double, double> MinMaxPressure(const Vector &u) const;
};

#endif // FWI_DGVACOUSTICASSEMBLECOMBINED_HPP
